package dp.decorator;

/**
 * Decorator pattern, an alternative inheritance
 * 
 * Instead of fixed inherited behavior we can dynamically
 * compose extended behavior
 * 
 * @author Wikipedia
 *
 */
public class Decorator {

    public static void main(String[] args) {
        // Create a decorated Window with horizontal and vertical scrollbars
        IWindow decoratedWindow = new HorizontalScrollBarDecorator(
                new VerticalScrollBarDecorator(new SimpleWindow()));

        // Print the Window's description
        System.out.println(decoratedWindow.getDescription());
    }

    // Interface
    interface IWindow {
        public void draw(); // draws the Window
        public String getDescription(); // returns a description of the Window
    }

    // implementation of a simple Window without any scrollbars
    static class SimpleWindow implements IWindow {
        public void draw() {
            // draw window
        }

        public String getDescription() {
            return "simple window";
        }
    }

    // abstract decorator class - note that it implements Window
    static abstract class WindowDecorator implements IWindow {
        protected IWindow decoratedWindow; // the Window being decorated

        public WindowDecorator(IWindow decoratedWindow) {
            this.decoratedWindow = decoratedWindow;
        }

        public void draw() {
            decoratedWindow.draw(); //delegation
        }

        public String getDescription() {
            return decoratedWindow.getDescription(); //delegation
        }
    }

    // the first concrete decorator which adds vertical scrollbar functionality
    static class VerticalScrollBarDecorator extends WindowDecorator {
        public VerticalScrollBarDecorator(IWindow decoratedWindow) {
            super(decoratedWindow);
        }

        @Override
        public void draw() {
            decoratedWindow.draw();
            drawVerticalScrollBar();
        }

        private void drawVerticalScrollBar() {
            // draw the vertical scrollbar
        }

        @Override
        public String getDescription() {
            return decoratedWindow.getDescription()
                    + ", including vertical scrollbars";
        }
    }

    // the second concrete decorator which adds horizontal scrollbar functionality
    static class HorizontalScrollBarDecorator extends WindowDecorator {
        public HorizontalScrollBarDecorator(IWindow decoratedWindow) {
            super(decoratedWindow);
        }

        @Override
        public void draw() {
            decoratedWindow.draw();
            drawHorizontalScrollBar();
        }

        private void drawHorizontalScrollBar() {
            // draw the horizontal scrollbar
        }

        @Override
        public String getDescription() {
            return decoratedWindow.getDescription()
                    + ", including horizontal scrollbars";
        }
    }

}
