package classinvariants;

import java.util.Date;
import java.util.Random;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

/**
 * Junit test for the Stack
 * 
 * @author hajo
 * 
 */
public class TestStack {

	// Used all over to check representation invariant.
	private boolean invariantHolds(Stack s) {
		return -1 <= s.getSp() && s.getSp() < s.size();
	}

	Stack s;

	@Before
	public void setUp() {
		s = new Stack();
	}

	@Test
	public void testPush() {
		Date d1 = new Date();

		Assert.assertTrue(s.size() == 0);
		Assert.assertTrue(invariantHolds(s));

		s.push(d1);
		s.push(d1);
		s.push(d1);

		// @post this.size.equals( \old(this.size) + 1 )
		Assert.assertTrue(s.size() == 3);
		Assert.assertTrue(invariantHolds(s));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testPushFail() {
		// @pre s != null
		s.push(null);
	}

	@Test
	public void testTop() {
		Date d1 = new Date();

		s.push(d1);
		int oldSize = s.size();
		Date d2 = s.top();
		int newSize = s.size();

		// this.size(\old(this.size))
		// \result == s => \old(this.push(s))
		Assert.assertTrue(oldSize == newSize);
		Assert.assertTrue(d2 == d1);
		Assert.assertTrue(invariantHolds(s));
	}

	@Test(expected = IllegalStateException.class)
	public void testTopFail() {
		// @pre this.size > 0
		s.top();
	}

	@Test
	public void testPop() {
		Date d1 = new Date();
		s.push(d1);
		Date d2 = s.top();
		Date d3 = s.pop();

		// @post \result == \old(this.top)
		// this.size == \old(this.size) - 1
		Assert.assertTrue(d2 == d3);
		Assert.assertTrue(s.size() == 0);
		Assert.assertTrue(invariantHolds(s));
	}

	@Test(expected = IllegalStateException.class)
	public void testPopFail() {
		Date d = new Date();
		s.push(d);
		s.push(d);
		s.pop();
		s.pop();
		// @pre this.size > 0
		s.pop();
	}

	@Test
	public void testPushResize() {
		Date d1 = new Date();

		// Will resize internal representation
		for (int i = 0; i < 9; i++) {
			s.push(d1);
		}
		for (int i = 0; i < 9; i++) {
			Date d2 = s.pop();
			Assert.assertTrue(d2 == d1);
		}
		Assert.assertTrue(invariantHolds(s));
		Assert.assertTrue(s.size() == 0);
	}

	@Test
	public void testSize() {
		Stack s = new Stack();
		Date d1 = new Date();
		Random r = new Random();

		int pushes = r.nextInt(10);
		int pops = pushes;

		for (int i = 0; i < 100; i++) {
			for (int j = 0; j < pushes; j++) {
				s.push(d1);
			}
			for (int j = 0; j < pops; j++) {
				s.pop();
			}
			// @post \result == | this.push* - this.pop* |
			Assert.assertTrue(s.size() == (pushes - pops));
		}
	}

	public void testRepExposure() {
		Stack s = new Stack();
		Date d1 = new Date();

		s.push(d1);

		long t1 = s.top().getTime();

		s.top().setTime(123456789); // Alter elem on top

		long t2 = s.top().getTime(); // Should give same as last top!

		Assert.assertFalse(t1 == t2);
	}

	public void testNoRepExposure() {
		Stack s = new Stack();
		Date d1 = new Date();

		s.push(d1);

		long t1 = s.topByValue().getTime();

		s.topByValue().setTime(123456789); // Alter a copy

		long t2 = s.topByValue().getTime();

		Assert.assertTrue(t1 == t2);
	}

}
